# awl.managment.commands.print_setting.py
#
# Takes the name of a django setting and prints the result to the screen. 

from django.conf import settings
from django.core.management.base import BaseCommand

class Command(BaseCommand):
    """Prints the named django setting to the screen. If the setting is a list
    or tuple, it prints each item in a single line separated by spaces so that
    it can be used as arguments to a command.  """

    def __init__(self, *args, **kwargs):
        super(Command, self).__init__(*args, **kwargs)
        self.help = self.__doc__

    def add_arguments(self, parser):
        parser.add_argument('name', type=str, 
            help='name of django setting to print')

    def handle(self, *args, **options):
        value = getattr(settings, options['name'])
        if isinstance(value, list) or isinstance(value, tuple):
            value = ' '.join(value)

        print(value)



# awl.decorators.py
import logging, json
from functools import wraps

from django.http import Http404

logger = logging.getLogger(__name__)

# ============================================================================
# View Decorators
# ============================================================================

def post_required(method_or_options=[]):
    """View decorator that enforces that the method was called using POST.
    This decorator can be called with or without parameters.  As it is
    expected to wrap a view, the first argument of the method being wrapped is
    expected to be a ``request`` object.

    .. code-block:: python

        @post_required
        def some_view(request):
            pass


        @post_required(['firstname', 'lastname'])
        def some_view(request):
            pass

    The optional parameter contains a single list which specifies the names of
    the expected fields in the POST dictionary.  The list is not exclusive,
    you can pass in fields that are not checked by the decorator.

    :param options:
        List of the names of expected POST keys.
    """
    def decorator(method):
        # handle wrapping or wrapping with arguments; if no arguments (and no
        # calling parenthesis) then method_or_options will be a list,
        # otherwise it will be the wrapped function
        expected_fields = []
        if not callable(method_or_options):
            # not callable means wrapping with arguments
            expected_fields = method_or_options

        @wraps(method)
        def wrapper(*args, **kwargs):
            request = args[0]
            if request.method != 'POST':
                logger.error('POST required for this url')
                raise Http404('only POST allowed for this url')

            missing = []
            for field in expected_fields:
                if field not in request.POST:
                    missing.append(field)

            if missing:
                s = 'Expected fields missing in POST: %s' % missing
                logger.error(s)
                raise Http404(s)

            # everything verified, run the view
            return method(*args, **kwargs)
        return wrapper

    if callable(method_or_options):
        # callable means decorated method without options, call our decorator 
        return decorator(method_or_options)
    return decorator


def json_post_required(*decorator_args):
    """View decorator that enforces that the method was called using POST and
    contains a field containing a JSON dictionary. This method should
    only be used to wrap views and assumes the first argument of the method
    being wrapped is a ``request`` object.

    .. code-block:: python

        @json_post_required('data', 'json_data')
        def some_view(request):
            username = request.json_data['username']

    :param field:
        The name of the POST field that contains a JSON dictionary
    :param request_name:
        [optional] Name of the parameter on the request to put the
        deserialized JSON data. If not given the field name is used

    """
    def decorator(method):
        @wraps(method)
        def wrapper(*args, **kwargs):
            field = decorator_args[0]
            if len(decorator_args) == 2:
                request_name = decorator_args[1]
            else:
                request_name = field

            request = args[0]
            if request.method != 'POST':
                logger.error('POST required for this url')
                raise Http404('only POST allowed for this url')

            if field not in request.POST:
                s = 'Expected field named %s in POST' % field
                logger.error(s)
                raise Http404(s)

            # deserialize the JSON and put it in the request
            setattr(request, request_name, json.loads(request.POST[field]))

            # everything verified, run the view
            return method(*args, **kwargs)
        return wrapper
    return decorator



from itertools import islice, chain
from django.db import models, transaction

from awl.absmodels import TimeTrackModel

# ============================================================================
# Concrete Models
# ============================================================================

class Counter(TimeTrackModel):
    """A named counter in the database with atomic update."""
    name = models.CharField(max_length=30)
    value = models.BigIntegerField(default=0)

    @classmethod
    def increment(cls, name):
        """Call this method to increment the named counter.  This is atomic on
        the database.

        :param name:
            Name for a previously created ``Counter`` object 
        """
        with transaction.atomic():
            counter = Counter.objects.select_for_update().get(name=name)
            counter.value += 1
            counter.save()

        return counter.value


class Lock(TimeTrackModel):
    """Implements a simple global locking mechanism across database accessors
    by using the ``select_for_update()`` feature.  Example:

    .. code-block:: python

        # run once, or use a fixture
        Lock.objects.create(name='everything')

        # views.py
        def something(request):
            Lock.lock_until_commit('everything')

    """
    name = models.CharField(max_length=30)

    @classmethod
    def lock_until_commit(cls, name):
        """Grabs this lock and holds it (using ``select_for_update()``) until
        the next commit is done.

        :param name:
            Name for a previously created ``Lock`` object 
        """
        Lock.objects.select_for_update().get(name=name)

# ============================================================================
# Misc
# ============================================================================

# Choices
#
# idea borrowed, modified then converted to python 3 from:
#  http://tomforb.es
#      /using-python-metaclasses-to-make-awesome-django-model-field-choices

class _ChoicesType(type):
    # Metaclass for the Choices object
    #

    def __new__(metacls, classname, bases, namespace, **kwds):
        # create the class
        cls = type.__new__(metacls, classname, bases, namespace, **kwds)

        if classname == 'Choices':
            # this is the abstract base class, do nothing more
            return cls

        if not hasattr(cls, '_choices_hash'):
            # define a has to contain our choices only if a parent hasn't done
            # it already -- this allows grandchildren of Choices to override
            # their parent's values
            cls._choices_hash = {}

        # values defined in the class (not in parent) are defined in __dict__,
        # loop through them and update _choices_hash; this will override
        # anything defined in a parent 
        for name in cls.__dict__.keys():
            value = getattr(cls, name)
            if name[0] == '_' or hasattr(value, '__call__') :
                # attribute is hidden or callable, ignore it

                # coverage is doing something funky with just a "continue"
                # here, ugly hack to force it to recognize the code
                x = 1
                x += 1
                continue

            if isinstance(value, tuple) and len(value) > 1:
                # value is a tuple, override the value to be just the first
                # part and set the content hash to be the second part
                setattr(cls, name, value[0])
                cls._choices_hash[value[0]] = value[1]
            else:
                # value is not a tuple, create a default choice name based on
                # the class attribute
                pieces = [x.capitalize() for x in name.split('_')]
                cls._choices_hash[value] = ' '.join(pieces)

        return cls

    def __iter__(cls):
        for name, value in cls._choices_hash.items():
            yield (name, value)


class Choices(metaclass=_ChoicesType):
    """
    .. note::
        Django 3.0 added `Enumeration Types <https://docs.djangoproject.com/en/3.0/ref/models/fields/#enumeration-types>`_
        which solves the same problem as this class. You should probably use
        it instead.

    A tuple of tuples pattern of ((id1, string1), (id2, string2)...)  is
    common in django for choices fields, etc.  This object inspects its own
    members (i.e. the inheritors) and produces the corresponding tuples.

    .. code-block:: python

        class Colours(Choices):
            RED = 'r'
            LIGHT_BLUE = 'b'

        >> Colours.RED
        'r'
        >> list(Colours)
        [('r', 'Red'), ('b', 'Light Blue')]

    A member value can also be a tuple to override the default string

    .. code-block:: python

        class Colours(Choices):
            RED = 'r'
            BLUE = 'b', 'Blueish'

        >> list(Colours)
        [('r', 'Red'), ('b', 'Blueish')]


    A inheriting class can also add or override members.

    .. code-block:: python

        class Colours(Choices):
            RED = 'r'
            BLUE = 'b'

        class MoreColours(Colours):
            GREEN = 'g'
            BLUE = 'b', 'Even More Blue'

        >> list(Colours)
        [('r', 'Red'), ('b', 'Even More Blue'), ('g', 'Green')]
    """
    @classmethod
    def get_value(cls, key):
        return cls._choices_hash[key]

# ----------------------------------------------------------------------------

# QuerySetChain
#
# borrowed and modified from:
#   http://stackoverflow.com/questions/431628/
#   by: http://stackoverflow.com/users/15770/akaihola
class QuerySetChain:
    """
    Chains together multiple querysets (possibly of different models) and 
    behaves as one queryset.  Supports minimal methods needed for use with
    django.core.paginator.  Does not support re-ordering or re-filtering
    across the set.

    .. code-block:: python

        q1 = Thing.objects.filter(foo)
        q2 = Stuff.objects.filter(bar)
        qsc = QuerySetChain(q1, q2)
    """

    def __init__(self, *subquerysets):
        self.querysets = subquerysets

    def count(self):
        """
        Performs a .count() for all subquerysets and returns the number of
        records as an integer.
        """
        return sum(qs.count() for qs in self.querysets)

    def _clone(self):
        "Returns a clone of this queryset chain"
        return self.__class__(*self.querysets)

    def _all(self):
        "Iterates records in all subquerysets"
        return chain(*self.querysets)

    def __getitem__(self, index):
        """
        Retrieves an item or slice from the chained set of results from all
        subquerysets.
        """
        if type(index) is slice:
            return list(islice(self._all(), index.start, index.stop, 
                index.step or 1))
        else:
            return next(islice(self._all(), index, index+1))



from django.db import models

# ============================================================================
# Abstract Models
# ============================================================================

class TimeTrackModel(models.Model):
    """Abstract model with auto-updating create & update timestamp fields.

    :param created:
        Date/time when the model was created
    :param updated:
        Date/time when the model was last updated
    """
    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)

    class Meta:
        abstract = True


class ValidatingMixin(models.Model):
    """Include this mixin to force model validation to happen on each
    ``save`` call.
    """
    class Meta:
        abstract = True

    def save(self, *args, **kwargs):
        self.full_clean()
        super(ValidatingMixin, self).save(*args, **kwargs)
